/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

/* #if USE_MMAP */
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
/* #endif */

#include "lxml.h"

/*! Structure referencing the XML file to read or to create
 */
struct LXMLhandle_s
{
  char *filename;   /**< the name of the file */
  int file;         /**< the file descriptor (set after lxml_file_open()) */
  int mode;         /**< the mode: LXML_Read or LXML_Write. */
  struct stat stat; /**< filled when opening a file for reading */
  char *encoding;   /**< the encoding of the file */
  char *dtd;        /**< the DTD associated to the file */
  char *map;        /**< pointer to the mmap'ed file */
  off_t idx;        /**< current offset in the mmap'ed file */
  off_t child;      
  off_t brother;
  int line;         /**< current line in the file */
  int depth;        /**< current tree depth */
};

/*! Structure representing an element in the XML tree */
struct LXMLelement_s
{
  char *name;       /**< the name of the element */
  int has_children; /**< whether the element has children */
  clist *attr;      /**< list of element attributes */
};

typedef struct LXMLattr_s LXMLattr;

/*! Structure representing an attribute */
struct LXMLattr_s
{
  char *key;        /**< the attribute name */
  char *value;      /**< the attribute value */
};

/* Declaration of static functions */

static int
lxml_set_string(char **dest, const char *string);

static int
lxml_getc(LXMLhandle *handle);

static int
lxml_isspace(int c);

static char *
lxml_parse_element_name(char *buf, int len, int *idx);

static cntree *
lxml_parse_next_element(LXMLhandle *handle);

static LXMLattr *
lxml_parse_attr(LXMLhandle *handle, char *buf, int len, int *idx);

static void
lxml_tree_free_cb(cntree *ntree, void *data);

static cntree *
lxml_file_read_r(LXMLhandle *handle);

static int
lxml_file_write_r(LXMLhandle *handle, cntree *ntree);

static int
lxml_file_write_prolog(LXMLhandle *handle);

static int
lxml_file_write_dtd(LXMLhandle *handle, const char *root_elem);

static int
lxml_write_buf(int fd, const char *string, size_t size);

static void
lxml_strcpy(char **buf, char *string);

static void
lxml_strcpy_escape(char **buf, char *string);

static int
lxml_tree_read(cntree *ntree, LXMLelement_callbacks *element_cb);

static int
lxml_tree_write(cntree *ntree, LXMLelement_callbacks *element_cb);

static LXMLelement *
lxml_element_new(void);

static void
lxml_element_free(LXMLelement *element, int free_attr_value);

static int
lxml_element_read_attr(LXMLelement *element, LXMLstore *store);

static int
lxml_element_write_attr(LXMLelement *element, LXMLstore *store);

static LXMLattr *
lxml_attr_new(char *name, int value_type, void *data);

static void
lxml_attr_free(LXMLattr *attr, int free_attr_value);

#define LXML_XML_VERSION_1_0 "xml version = \"1.0\""
#define LXML_INDENT_STRING "  "

/* Functions definitions */

/* XML <-> ntree conversion */

LXMLhandle
*lxml_new(void)
{
  LXMLhandle *handle;

  handle = (LXMLhandle *) malloc(sizeof(LXMLhandle));
  handle->filename = NULL;
  handle->file     = -1;
  handle->mode     = LXML_None;
  handle->encoding = NULL;
  handle->dtd      = NULL;
  handle->map      = NULL;
  handle->idx      = (off_t) 0;
  handle->child    = (off_t) 0;
  handle->brother  = (off_t) 0;
  handle->line     = 0;
  handle->depth    = 0;

  return handle;
}

void
lxml_free(LXMLhandle *handle)
{
  if (handle == NULL)
    return;
  if (handle->filename)
    free(handle->filename);  
  if (handle->encoding)
    free(handle->encoding);
  if (handle->dtd)
    free(handle->dtd);
  free(handle);
}

/* File parameters settings */

int
lxml_set_filename(LXMLhandle *handle, const char *filename)
{
  return lxml_set_string(&handle->filename, filename);
}

int
lxml_set_encoding(LXMLhandle *handle, const char *encoding)
{
  return lxml_set_string(&handle->encoding, encoding);
}

const char *
lxml_get_encoding(LXMLhandle *handle)
{
  return handle->encoding;
}

int
lxml_set_dtd(LXMLhandle *handle, const char *dtd)
{
  return lxml_set_string(&handle->dtd, dtd);
}

const char *
lxml_get_dtd(LXMLhandle *handle)
{
  return handle->dtd;
}

/*! Copy a string to another location, and free any previously stored string */
static int
lxml_set_string(char **dest, const char *string)
{
  if (*dest)
    free(*dest);

  if (string == NULL)
    *dest = NULL;
  else if ((*dest = strdup(string)) == NULL)
    {
      fprintf(stderr, "(LXML) set_string(): strdup(\"%s\"): %s\n", 
	      string, strerror(errno));
      return -1;
    }

  return 0;
}

/* File operations */

int
lxml_file_open(LXMLhandle *handle, int mode)
{
  int file;
  struct stat buf;
  char *map;

  handle->mode = mode;
  switch (mode)
    {
    case LXML_Read:
      file = open(handle->filename, O_RDONLY);
      break;
    case LXML_Write:
      file = open(handle->filename, O_WRONLY | O_CREAT | O_EXCL, 
		  S_IRUSR | S_IWUSR);
      break;
    default:
      fprintf(stderr, "(LXML) lxml_file_open(): unknown mode %d\n", mode);
      return -1;
    }
  
  if (file == -1)
    {
      fprintf(stderr, "%s: %s\n", handle->filename, strerror(errno));
      return -1;
    }
  
  handle->file = file;

  if (mode == LXML_Read)
    {
      if (fstat(file, &buf))
	{
	  perror("(LXML) lxml_file_open(): fstat()");
	  return -1;
	}
      map = (char *) mmap(NULL, buf.st_size, PROT_READ, MAP_PRIVATE, file, 0);
      if (map == MAP_FAILED)
	{
	  perror("(LXML) lxml_file_open(): mmap()");
	  return -1;
	}
      handle->map = map;
      handle->stat = buf;
      handle->line = 0;
    }

  return 0;
}

int
lxml_file_close(LXMLhandle *handle)
{
  int ret = 0;

  if (handle->file < 0)
    return 0;

  if (close(handle->file))
    {
      perror("(LXML) lxml_file_close(): close()");
      ret = -1;
    }
  if (handle->mode == LXML_Read && 
      munmap(handle->map, handle->stat.st_size))
    {
      perror("(LXML) lxml_file_close(): munmap()");
      ret = -1;
    }
  handle->map = NULL;
  handle->idx = 0;
  handle->file = -1;
  handle->mode = LXML_None;

  return ret;
}

/* Reading file */

cntree *
lxml_file_read(LXMLhandle *handle)
{
  handle->depth = 0;
  return lxml_file_read_r(handle);
}

/*! Read an XML file that was mmap'ed into memory.
 *  This function calls itself whenever an element
 *  has children
 */
static cntree *
lxml_file_read_r(LXMLhandle *handle)
{
  cntree *tree;
  LXMLelement *element;

  tree = lxml_parse_next_element(handle);

  if (tree && (element = (LXMLelement *) tree->data) && element->has_children)
    {
      cntree *child;

      handle->depth++;

      while ((child = lxml_file_read_r(handle)) != NULL)
	{
	  if (cntree_append_child(tree, child) == -1)
	    {
	      fprintf(stderr, "(LXML) lxml_file_read_r(): "
		      "cntree_append_child() failed\n");
	      break;
	    }
	}

      if (cntree_get_children_count(tree) == 0)
	element->has_children = 0;

      handle->depth--;
    }
  return tree;
}

/*! Version of getc to be used on mmap'ed regions instead of on files */
static int
lxml_getc(LXMLhandle *handle)
{
  int c;
  if (handle->idx >= handle->stat.st_size)
    return EOF;
  
  c = handle->map[handle->idx++];
  if (c == 0x0a)
      handle->line++;
  return c;
}

static int
lxml_isspace(int c)
{
  return (c == 0x20 || c == 0x09 || c == 0x0d || c == 0x0a);
}

/*! Return the first word in the buffer, ignoring leading spaces.
 *  \param buf the buffer where to search
 *  \param len maximal length of the buffer
 *  \param idx on entry: starting index for the search,
 *         on return, first space after the word
 *  \return a newly allocated string containing the word, or NULL
 *          if no word was found.
 */
static char *
lxml_parse_element_name(char *buf, int len, int *idx)
{
  int i = *idx;
  int j;
  char *name;

  while (i < len && buf[i] && lxml_isspace((int) buf[i]))
    i++;
  if (i >= len)
    {
      fprintf(stderr, "(LXML) element name not found\n");
      return NULL;
    }
  
  j = i + 1;
  while (j < len && buf[j] && !lxml_isspace((int) buf[j]))
    j++;
  
  name = (char *) malloc(j-i+1);
  if (name == NULL)
    {
      perror("(LXML) lxml_parse_element_name(): malloc()");
      return NULL;
    }
  name[j-i] = 0x00;
  memcpy(name, &buf[i], j - i);
  *idx = j;
  return name;
}

/*! Parse the next element in the XML mmap'ed file. This is a very basic
 *  parser. It checks for the first non space character. If it is not
 *  '<', then it returns an error. It then scans for '>', so '>' must
 *  be encoded in the attributes values. If the character after '<' is
 *  '!' (comment?) or '?' (document type declaration), the it
 *  skips it and checks for the next element.  If an element is found,
 *  then it checks if the opening tag started with '</', in which case
 *  it returns, because that tag was the end tag of a previous element.
 *  If it is indeed on opening tag, then it creates the LXMLelement and
 *  parses the attributes. If the opening tag did not closed with '/>', 
 *  then it sets the has_children flag of the element. No check is made
 *  if the closing tag matches the opening tag.
 */
static cntree *
lxml_parse_next_element(LXMLhandle *handle)
{
  int c, i, len;
  char *buf = NULL;
  char *name;
  int has_children;

  cntree *tree;
  LXMLelement *element;
  LXMLattr *attr;

  do
    {
      while ((c = lxml_getc(handle)) != EOF && lxml_isspace(c)) {}
      
      if (c == EOF)
	return NULL;
      
      if (c != '<')
	{
	  fprintf(stderr, "In file \"%s\" at line %d: "
		  "element start not found\n", 
		  handle->filename, handle->line);
	  return NULL;
	}
      
      i = 0;
      len = 8092;
      buf = (char *) realloc(buf, len);
      if (buf == NULL)
	{
	  perror("(LXML) lxml_parse_next_element(): realloc()");
	  return NULL;
	}
      
      while (i < len - 1 && ( c = lxml_getc(handle)) != EOF && 
	     c != '>') {
	buf[i] = (char) c;
	i++;
	if (i >= len - 1)
	  {
	    len *= 2;
	    buf = (char *) realloc(buf, len);
	    if (buf == NULL)
	      {
		perror("(LXML) lxml_parse_next_element(): realloc()");
		return NULL;
	      }
	  }
      }
      
      if (c == EOF)
	{
	  fprintf(stderr, "File \"%s\" ended unexpectedly.\n",
		  handle->filename);
	  free(buf);
	  return NULL;
	}
      
      if (i == 0)
	{
	  fprintf(stderr, "In file \"%s\" at line %d: "
		  "element is empty.\n",
		  handle->filename, handle->line);
	  free(buf);
	  return NULL;
	}
      
      buf[i] = 0x00;
      if (buf[0] == '/')
	{
	  free(buf);
	  return NULL;
	}
      has_children = buf[i-1] != '/';
      /* Handle comments that include other elements */

    } while (buf[0] == '!' || buf[0] == '?');
  
  i = 0;
  name = lxml_parse_element_name(buf, len, &i);
  
  if (!name)
    {
      fprintf(stderr, "In file \"%s\" at line %d: "
	      "error occured (see previous messag).\n",
	      handle->filename, handle->line);
      free(buf);
      return NULL;
    }

  element = lxml_element_new();
  if (element == NULL)
    {
      perror("(LXML) lxml_parse_next_element(): malloc()");
      free(name);
      free(buf);
      return NULL;
    }

  element->name = name;
  element->has_children = has_children;
  element->attr = clist_new();
  if (element->attr == NULL)
    {
      fprintf(stderr, "(LXML) lxml_parse_next_element(): clist_new() failed\n");
      free(buf);
      lxml_element_free(element, !NULL);
      return NULL;
    }
  tree = cntree_new();
  if (tree == NULL)
    {
      fprintf(stderr, "(LXML) lxml_parse_next_element(): cntree_new() failed\n");
      free(buf);
      lxml_element_free(element, !NULL);
      return NULL;
    }

  tree->data = element;

  while ((attr = lxml_parse_attr(handle, buf, len, &i)))
    {
      if (clist_append(element->attr, attr))
	{
	  fprintf(stderr, "(LXML) lxml_parse_next_element(): "
		  "clist_append() failed\n");
	  free(buf);
	  lxml_tree_free(tree); /* This frees the element, too */
	  return NULL;
	}
    }
  
  free(buf);

  return tree;
}

/*! Parse a buffer and create an attribute from the first pair
 *  of key=value strings.
 *  Parsing is very basic: it looks for the first non space
 *  character. If it is a NULL or '/' then it fails. No
 *  attributes names are expected to begin with '/'.
 *  It then scans for the first space or '=' character.
 *  Everything in between is considered the attribute name.
 *  So an attribute cannot have '=' character.
 *  It then advances to the first non space character that is
 *  equal to single or double quote. If no quotes are found
 *  and if it is not the end of the tag, then it is an error.
 *  If it is the end of the tag, then the attribute has a key
 *  but no value. If quotes are found, the matching quote
 *  is looked for and everything in between is the attribute
 *  value. If the attribute is parsed successfully, it is
 *  returned and the 'idx' is set so the next iteration of
 *  this function continues with the next attribute.
 */
static LXMLattr *
lxml_parse_attr(LXMLhandle *handle, char *buf, int len, int *idx)
{
  int i = *idx;
  int j, k;
  LXMLattr *attr;
  int esc;

  while (i < len && buf[i] && lxml_isspace((int) buf[i]))
    i++;
  if (buf[i] == 0x00 || buf[i] == '/' || i >= len)
    return NULL;
  j = i + 1;
  while (j < len && buf[j] && !lxml_isspace((int) buf[j]) && buf[j] != '=')
    j++;
  
  attr = (LXMLattr *) calloc(1, sizeof(LXMLattr));
  if (attr == NULL)
    {
      perror("(LXML) lxml_parse_attr(): malloc()");
      return NULL;
    }
  attr->key = (char *) malloc(j-i+1);
  if (attr == NULL)
    {
      perror("(LXML) lxml_parse_attr(): malloc()");
      free(attr);
      return NULL;
    }

  memcpy(attr->key, &buf[i], j - i);
  attr->key[j-i] = 0x00;
  attr->value = NULL;

  i = j + 1;
  while (i < len && buf[i] && lxml_isspace((int) buf[i]))
    i++;

  if (i >= len || buf[i] == 0x00)
    {
      *idx = i;
    return attr;
    }

  if (buf[i] != '"')
    {
      fprintf(stderr, "In file \"%s\" near line %d: "
	      "attribute value not quoted. Rest of attribute ignored.\n",
	      handle->filename, handle->line);
      lxml_attr_free(attr, 1);
      return NULL;
    }

  esc = 0;
  j = i + 1;
  while (j < len && buf[j] && buf[j] != '"')
    {
      if (buf[j] == '\\')
	{
	  j++;
	  esc++;
	  if (buf[j] == 0x00)
	    break;
	}
      j++;
    }
  
  if (buf[j] == 0x00)
    {
      fprintf(stderr, "In file \"%s\" near line %d: "
	      "mismatching quotes while looking for attribute's value.\n",
	      handle->filename, handle->line);
      free(attr->key);
      free(attr);
      return NULL;
    }

  attr->value = (char *) malloc(j - i - esc);
  if (attr->value == NULL)
    {
      perror("(LXML) lxml_parse_attr(): malloc()");
      free(attr->key);
      free(attr);
      *idx = j + 1;
      return NULL;
    }
  
  *idx = j + 1;

  /* Copy and Unescape */
  k = 0;
  for (j = i + 1; j < *idx - 1; j++)
    {
      if (buf[j] != '\\')
	attr->value[k++] = buf[j];
    }
  attr->value[k] = 0x00;

  return attr;
} 

/* Writing file */

int
lxml_file_write(LXMLhandle *handle, cntree *ntree)
{
  LXMLelement *element;

  handle->depth = 0;

  element = (LXMLelement *) ntree->data;

  lxml_file_write_prolog(handle);
  lxml_file_write_dtd(handle, element->name);

  return lxml_file_write_r(handle, ntree);
}

static int
lxml_file_write_r(LXMLhandle *handle, cntree *ntree)
{
  int r = -1;
  LXMLelement *element;
  LXMLattr *attr;
  clistiter *iter;
  size_t size = 0;
  size_t name_size;
  char *buf;
  char *tmp;
  int i;

  element = (LXMLelement *) ntree->data;
  
  /* Minimum size is indent_size + element name + 3, so the buffer
   * can always hold <element_name>, <element_name/> or </element_name>
   * To that, we add the \n (LF) line termination code
   */
  name_size = handle->depth * strlen(LXML_INDENT_STRING) + 
    strlen(element->name) + strlen("</>\n");

  size = name_size;

  if (element->attr)
    {
      
      for (iter = clist_begin(element->attr); iter; iter = iter->next)
	{
	  attr = (LXMLattr *) iter->data;
	  /* Count the space equal sign and the two quotes */
	  /* Double the count of attribute value to take into account
	   * the possible escaping of special characters
	   */
	  size += strlen(attr->key) + 2 * strlen(attr->value) + strlen(" =\"\"");
	}
    }

  buf = (char *) malloc(size + 1);

  if (buf)
    {
      tmp = buf;
      for (i = 0; i < handle->depth; i++)
	{
	  lxml_strcpy(&tmp, LXML_INDENT_STRING);
	}
      lxml_strcpy(&tmp, "<");
      lxml_strcpy(&tmp, element->name);
      if (element->attr)
	{
	  for (iter = clist_begin(element->attr); iter; iter = iter->next)
	    {
	      attr = (LXMLattr *) iter->data;
	      lxml_strcpy(&tmp, " ");
	      lxml_strcpy(&tmp, attr->key);
	      lxml_strcpy(&tmp, "=\"");
	      lxml_strcpy_escape(&tmp, attr->value);
	      lxml_strcpy(&tmp, "\"");
	    }
	}

      lxml_strcpy(&tmp, (element->has_children ? ">\n" : "/>\n"));
      r = lxml_write_buf(handle->file, buf, tmp - buf);
      free(buf);

      if (r == 0 && element->has_children && ntree->children)
	{
	  handle->depth++;
	  for (iter = clist_begin(ntree->children); 
	       iter; 
	       iter = iter->next)
	    {
	      lxml_file_write_r(handle, (cntree *) iter->data);
	    }
	  handle->depth--;
	  buf = (char *) malloc(name_size + 1);
	  if (buf)
	    {
	      tmp = buf;
	      for (i = 0; i < handle->depth; i++)
		{
		  lxml_strcpy(&tmp, LXML_INDENT_STRING);
		}
	      lxml_strcpy(&tmp, "</");
	      lxml_strcpy(&tmp, element->name);
	      lxml_strcpy(&tmp, ">\n");
	      r = lxml_write_buf(handle->file, buf, tmp - buf);
	      free(buf);
	    }
	  else
	    {
	      perror("(LXML) lxml_file_write(): malloc()");
	      r = -1;
	    }
	}
    }
  else
    {
      perror("(LXML) lxml_file_write(): malloc()");
      r = -1;
    }

  return r;  
}

static int
lxml_file_write_prolog(LXMLhandle *handle)
{
  int r;
  const char *prolog = "<?" LXML_XML_VERSION_1_0 "?>\n";
  const char *prolog_enc = "<?" LXML_XML_VERSION_1_0 " encoding=\"%s\" ?>\n";

  if (handle->encoding)
    {
      char *buf;
      buf = (char *) malloc(strlen(prolog_enc) + strlen(handle->encoding));
      if (buf)
	{
	  sprintf(buf, prolog_enc, handle->encoding);

	  r = lxml_write_buf(handle->file, buf, strlen(buf));
	  free(buf);
	}
      else
	{
	  perror("(LXML) lxml_file_write_prolog(): malloc()");
	  r = -1;
	}
    }
  else
    {
      r = lxml_write_buf(handle->file, prolog, strlen(prolog));
    }
  
  return r;
}

static int
lxml_file_write_dtd(LXMLhandle *handle, const char *root_elem)
{
  int r;
  const char *doctype = "<!DOCTYPE %s SYSTEM \"%s\">\n";
  char *buf;

  if (!root_elem || !handle->dtd)
    {
      return 0;
    }
  
  buf = (char *) malloc(strlen(doctype) + strlen(root_elem) + 
			strlen(handle->dtd));

  if (buf == NULL)
    {
      perror("(LXML) lxml_file_write_dtd(): malloc()");
      return -1;
    }
  
  sprintf(buf, doctype, root_elem, handle->dtd);

  r = lxml_write_buf(handle->file, buf, strlen(buf));

  free(buf);

  return r;
}

static int
lxml_write_buf(int fd, const char *string, size_t size)
{
  int r = 0;
  ssize_t total = 0;
  ssize_t count = 0;
  while (total < size && r == 0)
    {
      count = write(fd, string + total, size - total);
      if (count >= 0)
	{
	  total += count;
	}
      else
	{
	  perror("(LXML) lxml_write_buf(): write()");
	  r = -1;
	}
    }
  return r;
}

static void
lxml_strcpy(char **buf, char *string)
{
  for (;*string; (*buf)++, string++)
    {
      **buf = *string;
    }
  **buf = 0x00;
}

static void
lxml_strcpy_escape(char **buf, char *string)
{
  for (;*string; (*buf)++, string++)
    {
      if (string[0] == '"' || string[0] == '\\')
	{
	  **buf = '\\';
	  (*buf)++;
	}
      **buf = *string;
    }
  **buf = 0x00;
}

/* tree handling functions */

cntree *
lxml_tree_new(void)
{
  cntree *atree;
  LXMLelement *element;
  atree = cntree_new();
  element   = lxml_element_new();
  atree->data = element;
  return atree;
}

/* Free everything allocated for the N-tree */
void
lxml_tree_free(cntree *ntree)
{
  cntree_free(ntree, (lxml_tree_free_cb), (void *)1);
}

/* Free only the structural data: element names, attribute keys and
 * the cntree and the clist structures. The only data not freed
 * are the attribute values. The user should store pointers to the
 * attributes values so they are not lost.
 */
void
lxml_tree_free_structural_data(cntree *ntree)
{
  cntree_free(ntree, (lxml_tree_free_cb), NULL);
}

static void
lxml_tree_free_cb(cntree *ntree, void *data)
{
  LXMLelement *element;
  if (ntree == NULL)
    return;
  element = (LXMLelement *) ntree->data;
  lxml_element_free(element, (int )data);
}

LXMLelement *
lxml_tree_get_element(cntree* atree)
{
      return (LXMLelement *) cntree_get_data(atree);
}

/* tree <-> application data conversion */

int
lxml_tree_for_each_element_in_path(cntree *ntree, char **path,
			       int (*callback)(cntree *atree, 
					       void *adata, 
					       int aop), 
			       void *data,
			       int op)
{
  int r = 0;
  clistiter *iter;
  cntree *atree;
  LXMLelement *element;


  if (!ntree || !path[0])
    return 0;

  element = (LXMLelement *) ntree->data;
  if (element && element->name && strcmp(element->name, path[0]) == 0)
    {
      if (path[1] == NULL)
	return callback(ntree, data, op);

      if (ntree->children)
	{
	  for (iter = clist_begin(ntree->children); iter && r == 0; 
	       iter = iter->next)
	    {
	      atree = (cntree *) iter->data;
	      element = (LXMLelement *) atree->data;
	      if (element && element->name && 
		  strcmp(element->name, path[1]) == 0)
		r = lxml_tree_for_each_element_in_path(atree, path + 1, 
						       callback, data, op);
	    }
	}
    }
  return r;
}

int
lxml_tree_convert(cntree *ntree, LXMLelement_callbacks *element_cb, int op)
{
  if (element_cb == NULL)
    return 0;

  if (op == LXML_Read)
    return lxml_tree_read(ntree, element_cb);
  else
    return lxml_tree_write(ntree, element_cb);
}

static int
lxml_tree_read(cntree *ntree, LXMLelement_callbacks *element_cb)
{
  int r = 0;
  clistiter *iter;
  cntree *atree;
  LXMLelement *element;
  LXMLelement_callbacks *a_element_cb;
  void *data_iter;

  if (!ntree || !(ntree->children))
    return r;
      
  for (iter = clist_begin(ntree->children); iter; iter = iter->next)
    {
      atree = (cntree *) iter->data;
      if (atree && atree->data)
	{
	  element = (LXMLelement *) atree->data;
	  data_iter = NULL;
	  if (element->name)
	    for (a_element_cb = element_cb; 
		 r == 0 && a_element_cb && a_element_cb->element_name;
		 a_element_cb++)
	      {
		if (strcmp(element->name, a_element_cb->element_name) == 0)
		  r = a_element_cb->callback(atree,
					     a_element_cb->data, 
					     &data_iter,
					     LXML_Read);
	      }
	}
    }
  return r;
}

static int
lxml_tree_write(cntree *ntree, LXMLelement_callbacks *element_cb)
{
  int r = 0;
  cntree *atree;
  LXMLelement *element;
  LXMLelement_callbacks *a_element_cb;
  LXMLelement *nelement; 
  void *data_iter;


  if (!ntree)
    return r;
      
  nelement = lxml_tree_get_element(ntree);

  for (a_element_cb = element_cb; 
       r == 0 && a_element_cb && a_element_cb->element_name;
       a_element_cb++)
    {
      if (a_element_cb->data && *a_element_cb->data)
	{
	  data_iter = NULL;
	  do
	    {
	      atree = lxml_tree_new();
	      element = lxml_tree_get_element(atree);
	      lxml_element_name(element, a_element_cb->element_name, 
				LXML_Write);
	      r = a_element_cb->callback(atree, a_element_cb->data,
					 &data_iter, LXML_Write);
	      if (r == 0)
		{
		  nelement->has_children = 1;
		  cntree_append_child(ntree, atree);
		}
	      else
		lxml_tree_free(atree);
	    } while (r == 0 && data_iter);
	}
    }
  return r;
}

/* Element handling functions */

static LXMLelement *
lxml_element_new(void)
{
  LXMLelement *element;
  element = (LXMLelement *) malloc(sizeof(LXMLelement));
  if (element == NULL)
    return NULL;
  element->name = NULL;
  element->has_children = 0;
  element->attr = NULL;
  return element;
}

static void
lxml_element_free(LXMLelement *element, int free_attr_value)
{
  if (element == NULL)
    return;

  if (element->name)
    free(element->name);

  if (element->attr)
    {
      clistiter *cur;
      for (cur = clist_begin(element->attr); cur; cur = cur->next)
	lxml_attr_free((LXMLattr *) clist_content(cur), free_attr_value);
      clist_free(element->attr);
    }
  free(element);
}

int
lxml_element_name(LXMLelement *element, char *name, int op)
{
  if (op == LXML_Read)
    {
      return ((element->name && strcmp(element->name, name) == 0) ? 0 : -1);
    }
  else
    {
      element->name = strdup(name);
      return (element->name ? 0 : -1);      
    }
}

int
lxml_element_store_attr(LXMLelement *element, LXMLstore *store, int op)
{
  if (op == LXML_Read)
    return lxml_element_read_attr(element, store);
  else
    return lxml_element_write_attr(element, store);
}

static int
lxml_element_write_attr(LXMLelement *element, LXMLstore *store)
{
  LXMLattr *attr;
  clistiter *iter;
  int i;
  int r = 0;

  if (element == NULL || store == NULL)
    return -1;

  if (element->attr == NULL)
    element->attr = clist_new();

  for (i = 0; store[i].attr && r == 0; i++)
    {
      if (store[i].dest == NULL || *(store[i].dest) ==  NULL)
	continue;
      if (store[i].dest_kind == LXML_List)
	{
	  for (iter = clist_begin(((clist *) *(store[i].dest)));
	       iter;
	       iter = iter->next)
	    {
	      if (iter->data)
		{
		  attr = lxml_attr_new(store[i].attr, store[i].value_type,
					  iter->data);
		  if (attr)
		    clist_append(element->attr, attr);
		  else
		    r = -1;
		}
	    }
	}
      else
	{
	  attr = lxml_attr_new(store[i].attr, store[i].value_type,
				  *(store[i].dest));
	  if (attr)
	    clist_append(element->attr, attr);
	  else
	    r = -1;
	}
    }
  return r;
}

static int
lxml_element_read_attr(LXMLelement *element, LXMLstore *store)
{
  LXMLattr *attr;
  clistiter *iter;

  int i;

  if (element == NULL || store == NULL || element->attr == NULL)
    return -1;
  for (iter = clist_begin(element->attr); iter; iter = iter->next)
    {
      attr = (LXMLattr *) iter->data;
      if (attr == NULL || attr->key == NULL || attr->value == NULL)
	return -1;
      for (i = 0; store[i].attr; i++)
	{
	  if (strcmp(store[i].attr, attr->key) == 0)
	    {
	      if (store[i].value_type == LXML_String)
		{
		  if (store[i].dest_kind == LXML_List)
		    {
		      char *tmp;
		      clist *list;
		      tmp = strdup(attr->value);
		      if (tmp == NULL)
			return -1; /* strdup allocation pb */
		      if (*(store[i].dest) == NULL)
			*(store[i].dest) = clist_new();
		      list = (clist *) *(store[i].dest);
		      clist_append(list, tmp);
		    }
		  else
		    {
		      if (*(store[i].dest))
			{
			  if (store[i].dest_kind == LXML_Unique)
			    return -1; /* the attribute was not unique */
			  else 
			    free(*(store[i].dest));
			}
		      *(store[i].dest) = strdup(attr->value);
		      if (*(store[i].dest) == NULL)
			return -1; /* strdup allocation pb */
		    }
		}
	      else if (store[i].value_type == LXML_Int)
		{
		  int value = atoi(attr->value);
		  if (store[i].dest_kind == LXML_List)
		    {
		      clist *list;
		      if (*(store[i].dest) == NULL)
			*(store[i].dest) = clist_new();
		      list  = (clist *) *(store[i].dest);
		      clist_append(list, (void *) value);
		    }
		  else
		    *(store[i].dest) = (void *) value;
		}
	      else if (store[i].value_type == LXML_Boolean)
		{
		  int boolean = (strcmp(attr->value, "yes") == 0) || 
		    (strcmp(attr->value, "true") == 0);
		  if (store[i].dest_kind == LXML_List)
		    {
		      clist *list;
		      if (*(store[i].dest) == NULL)
			*(store[i].dest) = clist_new();
		      list = (clist *) *(store[i].dest);
		      clist_append(list, (void *) boolean);
		    }
		  else
		    *(store[i].dest) = (void *) boolean;
		}
	    }
	}
    }      
  return 0;
}

int
lxml_element_attr_count(LXMLelement *element)
{
  if (element->attr)
    return clist_count(element->attr);
  else
    return 0;
}

/* Attribute handling functions */

static LXMLattr *
lxml_attr_new(char *name, int value_type, void *data)
{
  LXMLattr *attr;

  attr = (LXMLattr *) malloc(sizeof(LXMLattr));
  if (attr)
    {
      attr->key = strdup(name);
      if (attr->key)
	{
	  switch (value_type)
	    {
	    case LXML_String:
	      attr->value = strdup((char *) data);
	      break;
	    case LXML_Boolean:
	      attr->value = strdup(( ((int) data) ? "yes" : "no"));
	      break;
	    case LXML_Int:
	      /* buf will contain a decimal representation of an int.
	       * An upper bound of the size of the decimal string
	       * would be log10(2^bits) = bits * log10(2)
	       * log10(2) being 0.3, and bits = 8*sizeof(int),
	       * 8 * 0.3 < 3. We add 1 for an eventual sign symbol.
	       * Hence the buffer size would be sizeof(int)*3 + 1
	       */
	      {
		char buf[sizeof(int)*3 + 1];
		sprintf(buf, "%d", ((int) data));
		attr->value = strdup(buf);
	      }
	      break;
	    default:
	      fprintf(stderr, "(LXML) lxml_attr_new(): "
		      "unknown attribute value type\n");
	      attr->value = NULL;
	    }
	  if (attr->value == NULL)
	    {
	      fprintf(stderr, "(LXML) lxml_attr_new(): "
		      "could not create attribute '%s'. Attribute ignored.\n",
		      name);
	      free(attr->key);
	      free(attr);
	      attr = NULL;
	    }
	}
      else
	{
	  perror("(LXML) lxml_attr_new(): strdup()");
	  free(attr->key);
	  free(attr);
	  attr = NULL;
	}
    }
  return attr;
}

static void
lxml_attr_free(LXMLattr *attr, int free_attr_value)
{
  if (attr == NULL)
    return;

  if (attr->key)
    free(attr->key);

  if (free_attr_value && attr->value)
    free(attr->value);

  free(attr);
}

/* vim: set cindent shiftwidth=2 tabstop=2 expandtab :*/  
