/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "intl.h"

#include "x_alloc.h"
#include "logging.h"

#include "folder.h"
#include "mailbox.h"


#define XEP_MAILBOX_USE_MULTIPLE_CONNECTIONS 1
#define XEP_MAILBOX_MAX_CONNECTIONS 10

typedef struct XePMailboxListCloneCbData_s XePMailboxListCloneCbData;

struct XePMailboxListCloneCbData_s
{
  XePMailboxListCloneCb user_cb;
  void *user_data;
  clist *mailboxes;
  unsigned int count;
};

typedef struct XePMailboxDisconnectData_s XePMailboxDisconnectData;
struct XePMailboxDisconnectData_s
{
  int count;
  XePMailbox *mailbox;
  XePMailboxDisconnectCb user_func;
  void *user_data;
};

typedef void (*XePMailboxCopyCb)(XePMailboxListCloneCbData *cb_data);

typedef struct XePMailboxCopyDataCb_s XePMailboxCopyDataCb;
struct XePMailboxCopyDataCb_s
{
  XePMailboxCopyCb user_cb;
  void *user_data;
};

/* Static global hash */

#define XEP_MAILBOX_HASH_SIZE 512

static cinthash_t *mailbox_uid_hash_table;

/* Functions declarations */

static void
Mailbox_free(XePMailbox *mailbox);

static void
Mailbox_copy(XePMailbox *source, 
	     XePMailbox *target,
	     XePMailboxCopyCb user_cb,
	     void *user_data);

static void
Mailbox_copy_finalize(void *acb_data);

static void
Mailbox_list_clone_free_finalize(XePMailbox *amailbox, 
				 void *acb_data);

static void
Mailbox_list_clone_finalize(XePMailboxListCloneCbData *acb_data);

static void
Mailbox_disconnect_all_cb(void *adata, void *afolder);

/* Functions definitions */

clist *
a_Mailbox_list_new(void)
{
  clist *mailboxes;

  mailboxes = clist_new();
  if (mailboxes == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("clist_new failed");
    }

  return mailboxes;
}

clist *
a_Mailbox_list_new_copy(clist *mailboxes)
{
  clist *mailboxes_copy;
  clistiter *iter;
  XePMailbox *mailbox;
  XePMailbox *mailbox_copy;
  int r = 0;

  mailboxes_copy = clist_new();

  if (mailboxes_copy == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));
      return NULL;
    }

  for (iter = clist_begin(mailboxes); iter && r == 0; iter = iter->next)
    {
      mailbox = (XePMailbox *) iter->data;
      mailbox_copy = a_Mailbox_new_copy(mailbox);
      if (mailbox_copy)
	{
	  if (clist_append(mailboxes_copy, mailbox_copy))
	    {
	      A_LOGGING_ERROR("malloc(): %s", strerror(errno));
	      a_Mailbox_free(mailbox_copy);
	      r = -1;
	    }
	}
      else
	{
	  r = -1;
	}
    }
  if (r == -1)
    {
      a_Mailbox_list_free_copy(mailboxes_copy);
      mailboxes_copy = NULL;
    }

  return mailboxes_copy;
}

int
a_Mailbox_list_clone(clist *mailboxes_copy,
		     XePMailboxListCloneCb user_cb,
		     void *user_data)
{
  clist *mailboxes;
  clistiter *iter;
  clistiter *next;
  XePMailbox *mailbox_copy;
  XePMailbox *mailbox;
  int r = 0;
  XePMailboxListCloneCbData *cb_data;

  cb_data = x_alloc(1, XePMailboxListCloneCbData);
  if (cb_data == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));
      user_cb(NULL, user_data);
      return -1;
    }


  mailboxes = clist_new();
  if (mailboxes == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));
      return -1;
    }

  cb_data->user_cb = user_cb;
  cb_data->user_data = user_data;
  cb_data->mailboxes = mailboxes;
  cb_data->count = 1;

  iter = clist_begin(mailboxes_copy);

  while (iter && r == 0)
    {
      mailbox_copy = (XePMailbox *) iter->data;
      next = iter->next;

      if (mailbox_copy->uid < 0)
	{
	  /* Mailbox marked for delete */
	  mailbox = a_Mailbox_get_from_uid(-mailbox_copy->uid);
	  if (mailbox)
	    {
	      cb_data->count++;
	      a_Mailbox_disconnect_all(mailbox,
				       Mailbox_list_clone_free_finalize,
				       cb_data);
	    }
	  else
	    {
	      A_LOGGING_SYSTEM_ERROR("attempt to delete a non-existent "
				     "mailbox %s", mailbox_copy->id);
	    }
	}
      else
	{
	  mailbox = a_Mailbox_get_from_uid(mailbox_copy->uid);
	  if (mailbox == mailbox_copy)
	    {
	      /* the registered mailbox is also the copy, that
	       * means, it is a newly created mailbox
	       * so use this registered copy as a real mailbox
	       * and create a copy that will replace it
	       * in the mailbox copies list
	       * All folders must already be registered
	       */
	      mailbox = a_Mailbox_new_copy(mailbox_copy);
	      iter->data = mailbox;
	      mailbox = mailbox_copy;
	      a_Folder_update_id(mailbox->folder);
	    }
	  else
	    {
	      /* The registered mailbox is not the copy,
	       * so the copy  is indeed a copy of an already 
	       * existing mailbox, just transfer the settings
	       * of the copy to the original one
	       * and disconnect deleted folders.
	       */
	      cb_data->count++;
	      Mailbox_copy(mailbox_copy, mailbox, 
			   Mailbox_list_clone_finalize,
			   cb_data);
	    }
	      
	  clist_append(mailboxes, mailbox);
	  
	}
      iter = next;
    }
  Mailbox_list_clone_finalize(cb_data);

  return r;
}

static void
Mailbox_list_clone_free_finalize(XePMailbox *mailbox, void *acb_data)
{
  a_Mailbox_free(mailbox);

  Mailbox_list_clone_finalize((XePMailboxListCloneCbData *) acb_data);
}

static void
Mailbox_list_clone_finalize(XePMailboxListCloneCbData *acb_data)
{
  XePMailboxListCloneCbData *cb_data = (XePMailboxListCloneCbData *) acb_data;
  XePMailboxListCloneCb user_cb;
  void *user_data;
  clist *mailboxes;

  cb_data->count--;
  if (cb_data->count)
    return;

  user_cb = cb_data->user_cb;
  user_data = cb_data->user_data;
  mailboxes = cb_data->mailboxes;

  free(cb_data);
  
  user_cb(mailboxes, user_data);  
}

void
a_Mailbox_list_free(clist *mailboxes)
{
  clistiter *iter;

  for (iter = clist_begin(mailboxes); iter; iter = iter->next)
    {
      XePMailbox *mailbox = (XePMailbox *) iter->data;
      a_Mailbox_free(mailbox);
    }
  clist_free(mailboxes);
}

void
a_Mailbox_list_free_copy(clist *mailboxes)
{
  clistiter *iter;

  for (iter = clist_begin(mailboxes); iter; iter = iter->next)
    {
      XePMailbox *mailbox = (XePMailbox *) iter->data;
      a_Mailbox_free_copy(mailbox);
    }
  clist_free(mailboxes);
}

static XePMailbox *
Mailbox_new(const char *id)
{
  XePMailbox *mailbox;

  mailbox = x_alloc(1, XePMailbox);
  if (mailbox == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));
      return NULL;
    }
  mailbox->id = strdup(id);
  if (mailbox->id == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("strdup(): %s", strerror(errno));
      free(mailbox);
      return NULL;
    }

  mailbox->name = NULL;
  mailbox->type = XEP_MAILBOX_UNKNOWN;
  mailbox->server = NULL;

  mailbox->inbox_ref = NULL;
  mailbox->draft_ref = NULL;
  mailbox->queue_ref = NULL;
  mailbox->sent_ref = NULL;

  mailbox->command = NULL;
  mailbox->cache = NULL;
  mailbox->flags = NULL;

  mailbox->use_multiple_connections = XEP_MAILBOX_USE_MULTIPLE_CONNECTIONS;
  mailbox->max_connections = XEP_MAILBOX_MAX_CONNECTIONS;

  mailbox->subscribed_only = 0;
  mailbox->folder = NULL;
  mailbox->folders = a_Folder_list_new();

  mailbox->ep_mail_list = NULL;
  mailbox->ep_mail_lsub = NULL;

  return mailbox;
}

XePMailbox *
a_Mailbox_new(const char *id)
{
  XePMailbox *mailbox;
  static long uid = 0;

  mailbox = Mailbox_new(id);
  if (mailbox)
    {
      mailbox->server = a_Server_new();
      /* TODO: remove the periodicity of mailbox uid, because in
       * the very long term, in theory, two mailboxes can have the
       * same uid
       */
      if (uid < 0) /* Use only the positive space */
	uid = 0;
      mailbox->uid = ++uid;
      if (mailbox_uid_hash_table == NULL)
	mailbox_uid_hash_table = cinthash_new(XEP_MAILBOX_HASH_SIZE);

      cinthash_add(mailbox_uid_hash_table, mailbox->uid, mailbox);
      /* TODO: register mailbox folders:
       */
      mailbox->folder = a_Folder_new(XEP_FOLDER_TYPE_MAILBOX);
      a_Folder_set_mailbox(mailbox->folder, mailbox);
      a_Folder_update_id(mailbox->folder);
    }
  return mailbox;
}

XePMailbox *
a_Mailbox_new_copy(XePMailbox *mailbox)
{
  XePMailbox *mailbox_copy;
  int r = 0;

  mailbox_copy = Mailbox_new(mailbox->id);

  if (mailbox_copy == NULL)
    return NULL;

  mailbox_copy->name      = X_STRDUP(mailbox->name);
  mailbox_copy->uid       = mailbox->uid;
  mailbox_copy->type      = mailbox->type;
  mailbox_copy->server    = a_Server_new_copy(mailbox->server);
  mailbox_copy->inbox_ref = X_STRDUP(mailbox->inbox_ref);
  mailbox_copy->draft_ref = X_STRDUP(mailbox->draft_ref);
  mailbox_copy->queue_ref = X_STRDUP(mailbox->queue_ref);
  mailbox_copy->sent_ref  = X_STRDUP(mailbox->sent_ref);
  mailbox_copy->command   = X_STRDUP(mailbox->command);
  mailbox_copy->cache     = X_STRDUP(mailbox->cache);
  mailbox_copy->flags     = X_STRDUP(mailbox->flags);

  mailbox_copy->use_multiple_connections = mailbox->use_multiple_connections;
  mailbox_copy->max_connections = mailbox->max_connections;

  mailbox_copy->subscribed_only = mailbox->subscribed_only;

  mailbox_copy->folder = a_Folder_new_copy(mailbox->folder);
  a_Folder_set_mailbox(mailbox_copy->folder, mailbox_copy);
  a_Folder_update_id(mailbox_copy->folder);

  a_Folder_list_free(mailbox_copy->folders);
  mailbox_copy->folders = a_Folder_list_new_copy(mailbox->folders);

  if (mailbox_copy->folders == NULL || r == -1)
    {
      a_Mailbox_free(mailbox_copy);
      mailbox_copy = NULL;
    }

  return mailbox_copy;
}

void
a_Mailbox_free(XePMailbox *mailbox)
{
  cinthash_remove(mailbox_uid_hash_table, mailbox->uid);
  a_Folder_free(mailbox->folder);
  a_Folder_list_free(mailbox->folders);
  Mailbox_free(mailbox);
}

void
a_Mailbox_free_copy(XePMailbox *mailbox)
{
  a_Folder_free_copy(mailbox->folder);
  a_Folder_list_free_copy(mailbox->folders);

  Mailbox_free(mailbox);
}

static void
Mailbox_free(XePMailbox *mailbox)
{
  x_free(mailbox->id);
  x_free(mailbox->name);

  a_Server_free(mailbox->server);

  x_free(mailbox->inbox_ref);
  x_free(mailbox->draft_ref);
  x_free(mailbox->queue_ref);
  x_free(mailbox->sent_ref);

  x_free(mailbox->command);
  x_free(mailbox->cache);
  x_free(mailbox->flags);

  if (mailbox->ep_mail_list)
    mail_list_free(mailbox->ep_mail_list);

  if (mailbox->ep_mail_lsub)
    mail_list_free(mailbox->ep_mail_lsub);

  x_free(mailbox);
}

/*! Make the target mailbox be a copy of the source mailbox
 *  It assumes the target mailbox is registered, and source folders
 *  are also registered
 *  The target will have:
 *  - identical id
 *  - identical name
 *  - it is assumed that both mailboxes have the same uid
 *  - the same type
 *  - identical server
 *  - identical default folders
 *  - identical command
 *  - identical cache and flags directories
 *  - identical MAILBOX folder copied by a_Folder_copy() with 
 *    folder's mailbox pointing to the target
 *  - identical folders list, copied by a_Folder_list_copy() with
 *    folders' mailbox pointing to the target
 */
static void
Mailbox_copy(XePMailbox *source, XePMailbox *target,
	     XePMailboxCopyCb user_cb,
	     void *user_data)
{
  XePMailboxCopyDataCb *cb_data;

  cb_data = x_alloc(1, XePMailboxCopyDataCb);
  if (cb_data == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));
      /* FIXME: use a return value to check for malloc failure */
      return;
    }
  cb_data->user_cb = user_cb;
  cb_data->user_data = user_data;

  x_free(target->id);
  target->id = strdup(source->id);

  x_free(target->name);
  target->name = X_STRDUP(source->name);

  /* the uid is already the same, the user being unable to change it */

  target->type = source->type;

  a_Server_copy(source->server, target->server);

  x_free(target->inbox_ref);
  target->inbox_ref = X_STRDUP(source->inbox_ref);

  x_free(target->draft_ref);
  target->draft_ref = X_STRDUP(source->draft_ref);

  x_free(target->queue_ref);
  target->queue_ref = X_STRDUP(source->queue_ref);

  x_free(target->sent_ref);
  target->sent_ref = X_STRDUP(source->sent_ref);

  x_free(target->command);
  target->command = X_STRDUP(source->command);

  x_free(target->cache);
  target->cache = X_STRDUP(source->cache);

  x_free(target->flags);
  target->flags = X_STRDUP(source->flags);

  target->use_multiple_connections = source->use_multiple_connections;
  target->max_connections = source->max_connections;

  target->subscribed_only = source->subscribed_only;

  /* FIXME: if the MAILBOX folder was a REAL folder (POP3, mbox)
   * then disconnect it if things have changed
   */
  a_Folder_copy(source->folder, target->folder);
  a_Folder_set_mailbox(target->folder, target);
  a_Folder_update_id(target->folder);

  a_Folder_list_copy(source->folders, target->folders, target,
		     Mailbox_copy_finalize, cb_data);
}

static void
Mailbox_copy_finalize(void *acb_data)
{
  XePMailboxCopyDataCb *cb_data = (XePMailboxCopyDataCb *) acb_data;
  XePMailboxCopyCb user_cb = cb_data->user_cb;
  void *user_data = cb_data->user_data;

  free(cb_data);
  user_cb(user_data);
}


void
a_Mailbox_set_type(XePMailbox *mailbox, XePMailboxType type)
{
  XePMailboxType old_type = mailbox->type;
  mailbox->type = type;
  if (type == XEP_MAILBOX_MBOX || type == XEP_MAILBOX_POP3)
    {
      mailbox->folder->type = XEP_FOLDER_TYPE_REAL;
      if (type == XEP_MAILBOX_POP3)
	{
	  if (a_Folder_get_path(mailbox->folder) || 
	      old_type == XEP_MAILBOX_POP3)
	    {
	      a_Folder_unregister(mailbox->folder);
	    }
	  a_Folder_update_id(mailbox->folder);
	  a_Folder_register(mailbox->folder);
	}
    }
  else
    mailbox->folder->type = XEP_FOLDER_TYPE_MAILBOX;
}

void
a_Mailbox_set_toplevel(XePMailbox *mailbox, const char *toplevel)
{
  if (a_Folder_get_path(mailbox->folder) || mailbox->type == XEP_MAILBOX_POP3)
    {
      a_Folder_unregister(mailbox->folder);
    }
  a_Folder_set_path(mailbox->folder, toplevel);
  a_Folder_update_id(mailbox->folder);
  a_Folder_register(mailbox->folder);
}      

const char *
a_Mailbox_get_id(XePMailbox *mailbox)
{
  return mailbox->id;
}

XePMailbox *
a_Mailbox_get_from_uid(long uid)
{
  return (XePMailbox *) cinthash_find(mailbox_uid_hash_table, uid);
}

const char *
a_Mailbox_type_as_string(XePMailboxType type)
{
  switch (type)
    {
    case XEP_MAILBOX_UNKNOWN:
      return "unknown";
      break;
    case XEP_MAILBOX_IMAP:
      return "imap";
      break;
    case XEP_MAILBOX_MBOX:
      return "mbox";
      break;
    case XEP_MAILBOX_MH:
      return "mh";
      break;
    case XEP_MAILBOX_NNTP:
      return "nntp";
      break;
    case XEP_MAILBOX_POP3:
      return "pop3";
      break;
    }
  return NULL;
}

XePMailboxType
a_Mailbox_type_from_string(const char *type)
{
  if (strcmp(type, "imap") == 0)
    return XEP_MAILBOX_IMAP;
  else if (strcmp(type, "mbox") == 0)
    return XEP_MAILBOX_MBOX;
  else if (strcmp(type, "mh") == 0)
    return XEP_MAILBOX_MH;
  else if (strcmp(type, "nntp") == 0)
    return XEP_MAILBOX_NNTP;
  else if (strcmp(type, "pop3") == 0)
    return XEP_MAILBOX_POP3;
  else
    return XEP_MAILBOX_UNKNOWN;
}

int
a_Mailbox_is_busy(XePMailbox *mailbox)
{
  clistiter *iter;
  XePFolder *folder;

  if (a_Folder_get_thread_state(mailbox->folder) == 
      XEP_MANAGER_THREAD_STATE_RUNNING)
    return -1;

  for (iter = clist_begin(mailbox->folders); iter; iter = iter->next)
    {
      folder = (XePFolder *) iter->data;
      if (a_Folder_get_thread_state(folder) == 
	  XEP_MANAGER_THREAD_STATE_RUNNING)
	return -1;
    }
  return 0;
}

void
a_Mailbox_disconnect_all(XePMailbox *mailbox, 
			 XePMailboxDisconnectCb user_func,
			 void *user_data)
{
  clistiter *iter;
  XePFolder *folder;
  XePMailboxDisconnectData *data;

  data = x_alloc(1, XePMailboxDisconnectData);
  data->count = 1 + clist_count(mailbox->folders);
  data->mailbox = mailbox;
  data->user_func = user_func;
  data->user_data = user_data;

  for (iter = clist_begin(mailbox->folders); iter; iter = iter->next)
    {
      folder = (XePFolder *) iter->data;
      a_Folder_disconnect_async(folder, Mailbox_disconnect_all_cb, data);
    }

  a_Mailbox_disconnect_async(mailbox, Mailbox_disconnect_all_cb, data);
}

static void
Mailbox_disconnect_all_cb(void *adata, void *afolder)
{
  XePMailboxDisconnectData *data = (XePMailboxDisconnectData *) adata;

  data->count--;

  if (data->count)
    return;

  if (data->user_func)
    data->user_func(data->mailbox, data->user_data);

  free(data);
}

void
a_Mailbox_disconnect_async(XePMailbox *mailbox,
 			   XePManagerMsgCb user_cb, void *user_data)
{
  XePFolder *folder = mailbox->folder;
  XePManagerMsg *msg;

  if (a_Folder_get_thread_state(folder) == XEP_MANAGER_THREAD_STATE_NONE)
    {
      if (user_cb)
	{
	  user_cb(user_data, folder);
	}
    }
  else
    {
      msg = a_Manager_msg_new();
      a_Manager_msg_set_type(msg, XEP_MANAGER_MAILBOX_DISCONNECT);
      a_Manager_msg_set_source_target(msg, NULL, &folder->thread);
      a_Manager_msg_set_args(msg, folder, folder);
      if (user_cb)
	{
	  a_Manager_msg_add_callback(msg, user_cb, user_data);
	}
      a_Manager_msg_submit(msg);
    }
}



