/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h> /* signal */

#if HAVE_UNAME
#  include <sys/utsname.h>
#endif

#include <sys/types.h>
#include <sys/wait.h>

#include <libetpan/libetpan.h> /* mailstream_debug */

#include "intl.h"

#include "logging.h"
#include "defines.h"
#include "x_alloc.h"
#include "xetpan.h"
#include "conf.h"
#include "ui.h"
#include "remote.h"
#include "manager.h"


/*! \mainpage XetPan
 * 
 * \section intro Introduction
 *
 * This is the internal documentation for the XetPan mail user agent.
 *
 */

/*! Verbosity level. Currently: 
 * 0 to be quiet,
 * 1 to be verbose, 
 * 2 to help debugging
 */
int verbose = 0;

/*! The different command line options.
 */
typedef enum
  {
    XETPAN_CLI_NONE          = 0,       /**< No option                     */
    XETPAN_CLI_COMPOSE       = 1 << 0,  /**< Compose a message.            */
    XETPAN_CLI_SEND          = 1 << 1,  /**< Send messages in queue.       */
    XETPAN_CLI_SEND_ALL_NOW  = 1 << 2,  /**< Send messages in queue, even 
					 *   if they were scheduled later. */
    XETPAN_CLI_RECEIVE       = 1 << 3,  /**< Retrieve messages.            */
    XETPAN_CLI_STATUS        = 1 << 4,  /**< Display number of messages.   */
    XETPAN_CLI_CONFIG        = 1 << 5,  /**< Set configuration directory.  */
    XETPAN_CLI_GTK           = 1 << 6,  /**< Use GTK+ user interface.      */
    XETPAN_CLI_TTY           = 1 << 7,  /**< Use ncurses, user interface.  */
    XETPAN_CLI_VIEW          = 1 << 8,  /**< View a RFC822 message         */
    XETPAN_CLI_VERBOSE       = 1 << 28, /**< Be verbose.                   */
    XETPAN_CLI_HELP          = 1 << 29, /**< Display help for CLI options. */
    XETPAN_CLI_VERSION       = 1 << 30, /**< Display version.              */
    XETPAN_CLI_ERROR         = 1 << 31  /**< Error: unrecognisable option. */
  } XePCLIFlag;

enum
  {
    XETPAN_OPT_SEARCH,
    XETPAN_OPT_FOUND,
    XETPAN_OPT_OK,
    XETPAN_OPT_NOTFOUND,
    XETPAN_OPT_ERROR,
    XETPAN_OPT_DONE,
    XETPAN_OPT_RETURN
  };

/*! Command line options structure.
 *  (From Dillo's CLI patch I submitted)
 */
typedef struct {
  char *sh_opt;           /**< Option's short name.  Must be non NULL. */
  char *lg_opt;           /**< Option's long name.  Must be non NULL.  */
  int opt_argc;           /**< Maximal Number of arguments.
                           *   If positive, all are mandatory; 
                           *   if negative, all are optional.          */
  XePCLIFlag opt_flg;     /**< Option flag.                            */
  char *help;             /**< Option's help string.  Can be NULL      */ 
} XePCLIOptions;


typedef struct
{
  XePCLIFlag type;
  char *arg;
} XePCmd;

XetPan *global_xetpan;

/*! Create the main data structure */
static XetPan *XetPan_new(void);

static void
XetPan_cmd_list_append(clist *cmd_list, XePCLIFlag opt_flg, char **opt_argv);

static void
XetPan_cmd_list_free(clist *cmd_list);

static void
XetPan_cmd_list_process(XetPan *xetpan, clist *cmd_list);

static XePCLIFlag
XetPan_get_opt(XePCLIOptions *xetpan_options, int argc, char **argv,
	       char ***opt_argv, int *cli_idx);

static int
XetPan_get_max_opt(XePCLIOptions *options);

/* Command line options functions */
static void
XetPan_print_version(void);

static void
XetPan_print_help(XePCLIOptions *xetpan_options);

static void
XetPan_print_builtin_ui(void);

static void
XetPan_receive_messages(XetPan *xetpan, char *account);

static void
XetPan_send_queued_messages(XetPan *xetpan, char *account);

static void
XetPan_send_all_now(XetPan *xetpan, char *account);

static void
XetPan_print_status(XetPan *xetpan, char *account);

static void
XetPan_request_remote_ui_tty(XetPan *xetpan);

static void
XetPan_remote_ui_tty(XePRemote *remote, char *terminal);

static void
XetPan_request_remote_ui_gtk(XetPan *xetpan, const char *terminal);

static void
XetPan_remote_ui_gtk(XePRemote *remote, char *option);

static
XetPan *XetPan_new(void)
{
  XetPan *xetpan;

  xetpan = x_alloc(1, XetPan);

  if (xetpan == NULL)
    {
      /* TODO: Debug output */
      A_LOGGING_ERROR("x_alloc: %s", strerror(errno));
      exit(-1);
    }

  xetpan->config_path     = NULL;

  xetpan->config          = NULL;
  xetpan->server_watch_id = 0;
  xetpan->server_fd       = 0;
  xetpan->ui              = a_UI_new();
  if (xetpan->ui == NULL)
    exit(-1);

  global_xetpan = xetpan;

  return xetpan;
}


static void
XetPan_free(XetPan *xetpan)
{
  if (!xetpan)
    return;

  if (xetpan->config_path)
    free(xetpan->config_path);
  a_Conf_free(xetpan->config);
  a_UI_free(xetpan->ui);
  free(xetpan);
  global_xetpan = NULL;
}

int
main(int argc, char **argv)
{
  XetPan *xetpan = NULL;
  int ret = 0;
  int quit = FALSE;
  unsigned int cli_options = 0; /* WARNING: must be 4 bytes unsigned int */
  int cli_idx;
  unsigned int opt_flg;
  char **opt_argv = NULL;
  int client;
  clist *cmd_list = NULL;

  XePCLIOptions xetpan_options[] = {
    {"-c", "--compose",      -1, XETPAN_CLI_COMPOSE,
     N_("  -c, --compose [EMAIL|MAILTO] Open a compose window.")},
    {"-s", "--send",         -1, XETPAN_CLI_SEND,
     N_("  -s, --send [ACCOUNT]         Send queued mails from ACCOUNT. If ACCOUNT\n"
	"                               is not present, use all accounts.")},
    {"-n", "--send-all-now", -1, XETPAN_CLI_SEND_ALL_NOW,
     N_("  -n, --send-all-now [ACCOUNT] Send all mails from ACCOUNT now, even if\n"
	"                               they were scheduled to be sent later.\n"
	"                               If ACCOUNT is not present, use all accounts")},
    {"-r", "--receive",      -1, XETPAN_CLI_RECEIVE,
     N_("  -r, --receive [ACCOUNT]      Receive mail from ACCOUNT. If ACCOUNT is\n"
	"                               not present, receive from all accounts.")},
    {"-S", "--status",       -1, XETPAN_CLI_STATUS,
     N_("  -S, --status [ACCOUNT]       Display new, unread and total numner of\n"
	"                               mails in account ACCOUNT. If ACCOUNT is\n"
	"                               not present, use all accounts.")},
    {"-m", "--message",       1, XETPAN_CLI_VIEW,
     N_("  -m, --message RFC822         Display the message contained in the\n"
	"                               file RFC822")},
    {"-C", "--config",        1, XETPAN_CLI_CONFIG,
     N_("  -C, --config CONFIG          Use CONFIG as configuration directory.")},
    {"-g", "--gtk",           -1, XETPAN_CLI_GTK,
     N_("  -g, --gtk [DISPLAY]          Use GTK+ GUI interface on DISPLAY.")},
    {"-t", "--tty",           0, XETPAN_CLI_TTY,
     N_("  -t, --tty                    Use TTY ncurses interface")},
    {"-V", "--verbose",       -1, XETPAN_CLI_VERBOSE,
     N_("  -V, --verbose                Verbosity level: 0, 1 or 2. Replaces\n"
	"                               environment variable XETPAN_VERBOSE")},
    {"-v", "--version",    0, XETPAN_CLI_VERSION,
     N_("  -v, --version                Display version info and exit.")},
    {"-h", "--help",       0, XETPAN_CLI_HELP,
     N_("  -h, --help                   Display this help text and exit.")},
    {NULL, NULL, 0, XETPAN_CLI_NONE, NULL}
  };

  verbose = (getenv("XETPAN_VERBOSE") == NULL ?
	     0 :
	     strtol(getenv("XETPAN_VERBOSE"), NULL, 0));

  setlocale(LC_ALL, "");
  signal(SIGPIPE, SIG_IGN);

#ifdef ENABLE_NLS
  bindtextdomain(PACKAGE, LOCALEDIR);
  bind_textdomain_codeset(PACKAGE, "UTF-8");
  textdomain(PACKAGE);
#endif

  xetpan = XetPan_new();

  client = a_Remote_check_for_server(xetpan);

  if (!client)
    cmd_list = clist_new();

  while (ret == 0)
    {
      opt_flg = XetPan_get_opt(xetpan_options, argc, argv, &opt_argv,
			       &cli_idx);

      if (opt_flg == 0)
	break;

      cli_options |= opt_flg;
      
      switch (opt_flg)
	{
	case XETPAN_CLI_HELP:
	  XetPan_print_help(xetpan_options);
	  quit = TRUE;
	  ret = 1;
	  break;
	case XETPAN_CLI_VERSION:
	  XetPan_print_version();
	  quit = TRUE;
	  ret = 1;
	  break;
	case XETPAN_CLI_CONFIG:
	  if (!client)
	    {
	      x_free(xetpan->config_path);
	      xetpan->config_path = strdup(opt_argv[0]);
	    }
	  break;
	case XETPAN_CLI_GTK:
	  if (client)
	    {
	      XetPan_request_remote_ui_gtk(xetpan, opt_argv[0]);
	    }
	  else
	    {
#ifdef USE_GTKUI             
	      xetpan->ui->type = XETPAN_UI_GTK;
	      if (opt_argv[0])
		{
		  x_free(xetpan->ui->terminal);
		  xetpan->ui->terminal = strdup(opt_argv[0]);
		}
#else
              printf(_("Error: Gtk+ User Interface not built in."));
              XetPan_print_builtin_ui();
              ret = 1;
              quit = TRUE;
#endif
	    }
          
	  break;
	case XETPAN_CLI_TTY:
	  if (client)
	    {
	      XetPan_request_remote_ui_tty(xetpan);
	    }
	  else
	    {
#ifdef USE_TTYUI
	      xetpan->ui->type = XETPAN_UI_TTY;
	      if (opt_argv[0])
		{
		  x_free(xetpan->ui->terminal);
		  xetpan->ui->terminal = strdup(opt_argv[0]);
		}
#else
              printf(_("Error: Tty User Interface not built in."));
              XetPan_print_builtin_ui();
              ret = 1;
              quit = TRUE;
#endif

	    }
	  break;
	case XETPAN_CLI_VIEW:
	  if (client)
	    {
	      A_LOGGING_ERROR("Viewing a message using a remote server"
			      "is not supported yet.");
#if 0
	      /* TODO: Implement viewing a message using an already 
	       * running instance
	       */
	      a_Remote_send_command(xetpan, 'm', opt_argv[0]);
#endif
	      quit = TRUE;
	    }
	  else
	    {
	      XetPan_cmd_list_append(cmd_list, opt_flg, opt_argv);
	    }
	  break;
	case XETPAN_CLI_VERBOSE:
	  if (opt_argv[0])
	    verbose = strtol(opt_argv[0], NULL, 0);
	  else
	    verbose = 1;
	  break;
	case XETPAN_CLI_RECEIVE:
	  if (client)
	    a_Remote_send_command(xetpan, 'r', opt_argv[0]);
	  else
	    XetPan_cmd_list_append(cmd_list, opt_flg, opt_argv);
	  break;
	case XETPAN_CLI_SEND:
	  if (client)
	    a_Remote_send_command(xetpan, 's', opt_argv[0]);
	  else
	    XetPan_cmd_list_append(cmd_list, opt_flg, opt_argv);
	  break;
	case XETPAN_CLI_SEND_ALL_NOW:
	  if (client)
	    a_Remote_send_command(xetpan, 'n', opt_argv[0]);
	  else
	    XetPan_cmd_list_append(cmd_list, opt_flg, opt_argv);
	  break;
	case XETPAN_CLI_STATUS:
	  if (client)
	    a_Remote_send_command(xetpan, 'S', opt_argv[0]);
	  else
	    XetPan_cmd_list_append(cmd_list, opt_flg, opt_argv);
	  break;
	case XETPAN_CLI_COMPOSE:
	  if (client)
	    {
	      a_Remote_send_command(xetpan, 'c', opt_argv[0]);
	      quit = TRUE;
	    }
	  else 
	    {
	      XetPan_cmd_list_append(cmd_list, opt_flg, opt_argv);
	    }
	  break;
	default:
	  /* X and Gtk options are ignored */
#if 0
	  ret = 1;
	  printf(_("Error in command line options.\n"));
	  return -1;
#endif
	  break;		 
	}
    }
  
  if (quit || client)
    {
      A_LOGGING_INFO("End of client.");
      XetPan_free(xetpan);
      exit(0);
    }

  xetpan->config = a_Conf_read(xetpan->config_path);

  /* Initialize threads manager */
  a_Manager_init(xetpan);

  /* Initialize User Interface */
  if (a_UI_init(xetpan, &argc, &argv))
    exit(1);

  if (verbose)
    mailstream_debug = 1;

#if 0
  /* FIXME: a lockup could happen if loggin is too verbose 
   */
  if ((ret = a_Logging_redirect(xetpan, xetpan->config_path)) != 0)
    {
#if 0
       a_UI_error(_("System error"), _("Logging will not be viewable in"
				       "the log window")); 
#endif
    }
#endif
  if (!a_Remote_be_a_server(xetpan))
    {
      /* TODO: Debug output */
      /* Error message. A pointer that will be used later
       * to output to the console or on a GUI 
       */
    }

  A_LOGGING_INFO("Switched to server mode");

  if (!xetpan->config)
    {
      XetPan_cmd_list_free(cmd_list);
      a_UI_first_launch(xetpan);
    }
  else 
    {
      XetPan_cmd_list_process(xetpan, cmd_list);

      if ((cli_options & (XETPAN_CLI_COMPOSE | XETPAN_CLI_VIEW)) == 0)
	{
	  a_UI_mainview_create(xetpan, xetpan->ui->terminal, xetpan->ui->type);
	}
    }

  XetPan_cmd_list_free(cmd_list);

  /* Main loop */  
  a_UI_main(xetpan);

  a_Manager_quit(xetpan);
  XetPan_free(xetpan);
  exit(0);
}

static void
XetPan_cmd_list_append(clist *cmd_list, XePCLIFlag opt_flg, char **opt_argv)
{
  XePCmd *cmd;

  cmd = x_alloc(1, XePCmd);
  cmd->type = opt_flg;
  cmd->arg = (opt_argv[0] ? strdup(opt_argv[0]) : NULL);

  clist_append(cmd_list, cmd);
}

static void
XetPan_cmd_list_free(clist *cmd_list)
{
  clistiter *iter;
  XePCmd *cmd;
  for (iter = clist_begin(cmd_list); iter; iter = iter->next)
    {
      cmd = (XePCmd *) iter->data;
      x_free(cmd->arg);
      free(cmd);
    }
  clist_free(cmd_list);
}

static void
XetPan_cmd_list_process(XetPan *xetpan, clist *cmd_list)
{
  clistiter *iter;
  XePCmd *cmd;
  XePUIMessageView *mview;
  XePUIComposeWindow *uicompose;

  for (iter = clist_begin(cmd_list); iter; iter = iter->next)
    {
      cmd = (XePCmd *) iter->data;
      switch (cmd->type)
	{
	case XETPAN_CLI_VIEW:
	  mview = a_UI_messageview_create(xetpan, 
					  xetpan->ui->terminal,
					  xetpan->ui->type);
	  if (mview)
	    {
	      a_UI_messageview_view_file(mview, cmd->arg);
	    }
	  break;
	case XETPAN_CLI_RECEIVE:
	  XetPan_receive_messages(xetpan, cmd->arg);
	  break;
	case XETPAN_CLI_SEND:
	  XetPan_send_queued_messages(xetpan, cmd->arg);
	  break;
	case XETPAN_CLI_SEND_ALL_NOW:
	  XetPan_send_all_now(xetpan, cmd->arg);
	  break;
	case XETPAN_CLI_STATUS:
	  XetPan_print_status(xetpan, cmd->arg);
	  break;
	case XETPAN_CLI_COMPOSE:
	  uicompose = a_UI_composewindow_compose(xetpan, 
						 xetpan->ui->terminal,
						 xetpan->ui->type,
						 cmd->arg);
	  break;
	default:
	  printf("Unknown command line option %d\n", cmd->type);
	}
    }
}

/* Command line options functions */

/*! Display version and copyright information - Command Line Function. */
static void
XetPan_print_version(void)
{
#if HAVE_UNAME
  struct utsname utsbuf;
  uname(&utsbuf);
#endif
  printf("\n");
  printf(_("XetPan, version %s running on "), VERSION);

#if HAVE_UNAME
  printf("%s %s (%s)", utsbuf.sysname, utsbuf.release, utsbuf.machine);
#else
  printf (_("an unknown system."));
#endif

  printf(_("\nCompiled with libEtPan! %d.%d\n"),
	 LIBETPAN_VERSION_MAJOR, 
	 LIBETPAN_VERSION_MINOR);
  printf(_("Linked with libEtPan! %d.%d\n"),
	 libetpan_get_version_major(), 
	 libetpan_get_version_minor());
  printf ("\n");
  printf (_("Copyright (C) 2002, 2003 Melvin Hadasht.\n"
            "This is free software with "
	    "ABSOLUTELY NO WARRANTY.\n\n"));
}

/*! Display help text about command line options. 
 *  \param xetpan_options a pointer to the options structure.
 */
static void
XetPan_print_help(XePCLIOptions *xetpan_options)
{
  printf(_("\nUsage: xetpan [OPTIONS]\n"
	   "Options:\n"));
  
  for (; xetpan_options && xetpan_options->sh_opt; xetpan_options++)
    if (xetpan_options->help)
      printf("%s\n", gettext(xetpan_options->help));
    else
      printf(_("  %s, %s   *Undocumented*\n"), xetpan_options->sh_opt,
	     xetpan_options->lg_opt);

  XetPan_print_builtin_ui();
}

/*! Display built in User Interfaces 
 */
static void
XetPan_print_builtin_ui(void)
{
  printf(_("\nBuilt-in User Interfaces:\n"));
#ifdef USE_GTKUI
  printf(_("gtk: Gtk+ Graphical User Interface\n"));
#endif
#ifdef USE_TTYUI
  printf(_("tty: terminal (ncurses) User Interface\n"));
#endif
  printf("\n");
}
/*! Retrieve messages.
 *  \param xetpan the main XetPan structure.
 *  \param account the account name.
 */
static void
XetPan_receive_messages(XetPan *xetpan, char *account)
{
  A_LOGGING_INFO("Receiving from %p account %s", xetpan, account);
}

/*! Send queued messages.
 * Message that are scheduled to be sent later won't be sent.
 *  \param xetpan the main XetPan structure.
 *  \param account the account name. NULL means all accounts..
 */
static void
XetPan_send_queued_messages(XetPan *xetpan, char *account)
{
  A_LOGGING_INFO("Sending %p from account %s", xetpan, account);
}

/*! Send all queued messages.
 * Even message that are scheduled to be sent later will be sent.
 *  \param xetpan the main XetPan structure.
 *  \param account the account name. NULL means all accounts.
 */
static void
XetPan_send_all_now(XetPan *xetpan, char *account)
{
  A_LOGGING_INFO("Sending all now %p from account %s", xetpan, account);
}

/*! Display new, unread and total mail number in account/mailbox/folder. 
 *  \param xetpan the main XetPan structure.
 *  \param account account/mailbox/folder name.
 */
static void
XetPan_print_status(XetPan *xetpan, char *account)
{
  A_LOGGING_INFO("Printing status %p from account %s", xetpan, account);
}

static void
XetPan_request_remote_ui_tty(XetPan *xetpan)
{
  char *term;
  char *tty;
  char *argument;

  int len = 0;

  term = getenv("TERM");
  if (!term)
    {
      A_LOGGING_ERROR("TERM environment variable not set. "
		      "Running as a remote client impossible.");
      return;
    }

  len = strlen(term);

  tty = ttyname(0);
  if (!tty)
    {
      A_LOGGING_ERROR("Not running from a terminal. "
		      "Running as a remote client impossible.");
      return;
    }

  len += strlen(tty);

  len++; /* tab separator */
  len++; /* terminating NULL character */


  argument = x_alloc(len, char);
  if (!argument)
    {
      A_LOGGING_ERROR("x_alloc: %s", strerror(errno));
      return;
    }

  snprintf(argument, len, "%s\t%s", term, tty);

  a_Remote_send_command(xetpan, 't', argument);  

  x_free(argument);
}

static void
XetPan_remote_ui_tty(XePRemote* remote, char *terminal)
{
  XePUIMain *uimain;

#ifdef USE_TTYUI
  uimain = a_UI_mainview_create(remote->xetpan, terminal, XETPAN_UI_TTY);

  if (uimain)
    {
      uimain->remote = remote;
      A_LOGGING_INFO("TTY UI has remote connection fd %d", remote->fd);
    }
  
  a_Remote_build_msg(remote, 't', "OK.");
#else
  a_Remote_build_msg(remote, 't', "ERROR: Tty User Interface not built in.");
#endif
}

static void
XetPan_request_remote_ui_gtk(XetPan *xetpan, const char *terminal)
{
  const char *argument;

  if (terminal)
    argument = terminal;
  else 
    argument = getenv("DISPLAY");

  if (!argument)
    {
      A_LOGGING_ERROR("Display not specified and DISPLAY environment "
		      "variable unset. Exporting windows impossible.");
    }

  a_Remote_send_command(xetpan, 'g', argument);  
}

static void
XetPan_remote_ui_gtk(XePRemote *remote, char *option)
{
#ifdef USE_GTKUI
  if (remote->xetpan->ui->type != XETPAN_UI_GTK)
    a_UI_main_switch(remote->xetpan, option, XETPAN_UI_GTK);
  else
    a_UI_mainview_create(remote->xetpan, option, XETPAN_UI_GTK);

  a_Remote_build_msg(remote, 'g', "OK.");
#else
  a_Remote_build_msg(remote, 'g', "ERROR: Gtk+ User Interface not built in.");
#endif
}

int
a_XetPan_execute_remote_cmd(XePRemote *remote, char cmd, char *option)
{
  XetPan *xetpan;

  int ret;
  
  /*  if (xetpan->config == NULL)
    {
      printf(_("Received a remote command, but no configuration is available, "
	       "yet.\nCommand ignored.\n"));
      return XETPAN_ERR_UNKNWN_CMD;
    }
  */

  A_LOGGING_INFO("Executing remote command '%c' with option '%s'", cmd, option);
 
  xetpan = remote->xetpan;

  ret = XETPAN_NO_ERR;
  switch (cmd)
    {
    case 'r':
      XetPan_receive_messages(xetpan, option);
      a_Remote_build_msg(remote, cmd, option);
      break;
    case 's':
      XetPan_send_queued_messages(xetpan, option);
      a_Remote_build_msg(remote, cmd, option);
      break;
    case 'n':
      XetPan_send_all_now(xetpan, option);
      a_Remote_build_msg(remote, cmd, option);
      break;
    case 'S':
      XetPan_print_status(xetpan, option);
      a_Remote_build_msg(remote, cmd, option);
      break;
    case 'c':
      a_UI_composewindow_compose(xetpan, 
				 xetpan->ui->terminal,
				 xetpan->ui->type,
				 option);
      a_Remote_build_msg(remote, cmd, option);
      break;
    case 't':
      XetPan_remote_ui_tty(remote, option);
#ifdef USE_TTYUI
      ret = XETPAN_REMOTE_CLIENT;
#endif
      break;
    case 'g':
      XetPan_remote_ui_gtk(remote, option);
      break;
    default:
      a_Remote_build_msg(remote, cmd, option);
      ret = XETPAN_ERR;
    }
  A_LOGGING_INFO("Sending command result to remote instance");

  return ret;
}

static int
XetPan_get_max_opt(XePCLIOptions *options)
{
  int i, max;

  for (i = 0, max = 0; options[i].sh_opt; i++)
    if (abs(options[i].opt_argc) > max)
      max = abs(options[i].opt_argc);

  return max;
}

/*! Command line parser.
 *  Routine submitted to Dillo and enhanced by Jorge Arellano Cid.
 *  \param xetpan_options a pointer to the options structure.
 *  \param argc number of command line arguments.
 *  \param argv command line arguments array.
 *  \param opt_argv pointer to the option's argument array (NULL terminated).
 *         Pass NULL for the first call. Freeed by the function when no more
 *         options are available.
 *  \param cli_idx index for the arguments array.
 *  \return the command line flag corresponding to the option.
 */
static XePCLIFlag
XetPan_get_opt(XePCLIOptions *options, int argc, char **argv,
	       char ***opt_argv, int *cli_idx)
{
   static int idx = 1;
   int i = 0;
   int n_arg;
   int state;
   XePCLIFlag opt_flg = XETPAN_CLI_NONE;
   XePCLIFlag ret_val = XETPAN_CLI_NONE;;

   /* Create a sufficiently large array to hold the maximum possible number of
    * option arguments */
   if (*opt_argv == NULL)
     *opt_argv = x_alloc(XetPan_get_max_opt(options) + 1, char *);

   state = XETPAN_OPT_SEARCH;
   if (idx >= argc)
       state = XETPAN_OPT_DONE;
   
   /* Find the option */
   while (state != XETPAN_OPT_RETURN)
     {
       switch (state)
	 {
	 case XETPAN_OPT_SEARCH:
	   for (i = 0; options[i].sh_opt; i++)
	     if (strcmp(options[i].sh_opt, argv[idx]) == 0 ||
		 strcmp(options[i].lg_opt, argv[idx]) == 0)
	       break;
	   state = (options[i].sh_opt) ? XETPAN_OPT_FOUND : 
	     XETPAN_OPT_NOTFOUND;
	   break;
	 case XETPAN_OPT_FOUND:
	   idx++;
	   opt_flg = options[i].opt_flg;
	   n_arg = options[i].opt_argc;
	   /* Find the required/optional arguments of the options */
	   for (i = 0;
		idx < argc && i < abs(n_arg) && argv[idx][0] != '-'; 
		i++)
	     {
	       (*opt_argv)[i] = argv[idx++];
	     }
	   (*opt_argv)[i] = NULL;

	   /* Optional arguments have opt_argc < 0 */
	   if (i < n_arg)
	     {
	       printf(_("Option %s requires %d argument(s)\n"), 
		      argv[idx - i - 1], n_arg);
	       state = XETPAN_OPT_ERROR;
	     }
	   else
	     {
	       state = XETPAN_OPT_OK;
	     }
	   break;
	 case XETPAN_OPT_OK:
	   *cli_idx = idx;
	   ret_val = opt_flg;
	   state = XETPAN_OPT_RETURN;
	   break;
	 case XETPAN_OPT_NOTFOUND:
	   ret_val = XETPAN_CLI_NONE;
	   state = XETPAN_OPT_DONE;
	   if (strcmp(argv[idx], "--") == 0)
	     idx++;
	   else if (argv[idx][0] == '-')
	     state = XETPAN_OPT_ERROR;
	   break;
	 case XETPAN_OPT_ERROR:
	   ret_val = XETPAN_CLI_ERROR;
	   state = XETPAN_OPT_DONE;
	   break;
	 case XETPAN_OPT_DONE:
	   x_free(*opt_argv);
	   *cli_idx = idx;
	   state = XETPAN_OPT_RETURN;
	   break;
	 }
     }

   return ret_val;
}

int
a_XetPan_should_end(XetPan *xetpan)
{
  int ret;
  if (!xetpan)
    return -1;
  
  ret = ((xetpan->ui->mainviews->count == 0) &&
	 clist_isempty(xetpan->ui->composewindows));

  A_LOGGING_DEBUG("should end = %d", ret);

  return ret;
}

void
a_XetPan_quit(XetPan *xetpan)
{
  int uitype;
  A_LOGGING_TRACE;

  for (uitype = 0; uitype < XETPAN_UI_NB; uitype++)
    a_UI_close_all(xetpan, uitype);

  a_UI_main_quit(xetpan);
  
}
